from typing import List
from typing import Optional
from typing import Union

from superannotate_schemas.schemas.base import BaseVectorInstance
from superannotate_schemas.schemas.base import BboxPoints
from superannotate_schemas.schemas.base import Comment
from superannotate_schemas.schemas.base import BaseImageMetadata
from superannotate_schemas.schemas.base import Tag
from superannotate_schemas.schemas.base import BaseAttribute
from superannotate_schemas.schemas.base import AxisPoint
from superannotate_schemas.schemas.enums import VectorAnnotationTypeEnum
from superannotate_schemas.schemas.base import StrictNumber
from superannotate_schemas.schemas.base import NotEmptyStr

from pydantic import BaseModel
from pydantic import StrictInt
from pydantic import StrictFloat
from pydantic import conlist
from pydantic import Field
from pydantic import StrictStr
from pydantic import validate_model
from pydantic import ValidationError
from pydantic import validator


class Attribute(BaseAttribute):
    name: NotEmptyStr
    group_name: NotEmptyStr = Field(alias="groupName")


class VectorInstance(BaseVectorInstance):
    attributes: Optional[List[Attribute]] = Field(list())


class Metadata(BaseImageMetadata):
    is_predicted: Optional[bool] = Field(alias="isPredicted")


class Point(VectorInstance, AxisPoint):
    pass


class PolyLine(VectorInstance):
    points: List[Union[StrictFloat, StrictInt]]


class Polygon(VectorInstance):
    points: conlist(Union[StrictFloat, StrictInt], min_items=3)


class Bbox(VectorInstance):
    points: BboxPoints


class RotatedBoxPoints(VectorInstance):
    x1: StrictNumber
    y1: StrictNumber
    x2: StrictNumber
    y2: StrictNumber
    x3: StrictNumber
    y3: StrictNumber
    x4: StrictNumber
    y4: StrictNumber


class RotatedBox(VectorInstance):
    points: RotatedBoxPoints


class Ellipse(VectorInstance):
    cx: StrictNumber
    cy: StrictNumber
    rx: StrictNumber
    ry: StrictNumber
    angle: StrictNumber


class TemplatePoint(BaseModel):
    id: StrictInt
    x: StrictNumber
    y: StrictNumber


class TemplateConnection(BaseModel):
    id: StrictInt
    from_connection: StrictInt = Field(alias="from")
    to_connection: StrictInt = Field(alias="to")


class Template(VectorInstance):
    points: conlist(TemplatePoint, min_items=1)
    connections: List[TemplateConnection]
    template_id: Optional[StrictInt] = Field(None, alias="templateId")
    template_name: StrictStr = Field(alias="templateName")


class CuboidPoint(BaseModel):
    f1: AxisPoint
    f2: AxisPoint
    r1: AxisPoint
    r2: AxisPoint


class Cuboid(VectorInstance):
    points: CuboidPoint


ANNOTATION_TYPES = {
    VectorAnnotationTypeEnum.BBOX: Bbox,
    VectorAnnotationTypeEnum.TEMPLATE: Template,
    VectorAnnotationTypeEnum.CUBOID: Cuboid,
    VectorAnnotationTypeEnum.POLYGON: Polygon,
    VectorAnnotationTypeEnum.POINT: Point,
    VectorAnnotationTypeEnum.POLYLINE: PolyLine,
    VectorAnnotationTypeEnum.ELLIPSE: Ellipse,
    VectorAnnotationTypeEnum.RBBOX: RotatedBox,
}


class VectorAnnotation(BaseModel):
    metadata: Metadata
    comments: Optional[List[Comment]] = Field(list())
    tags: Optional[List[Tag]] = Field(list())
    instances: Optional[
        List[
            Union[Template, Cuboid, Point, PolyLine, Polygon, Bbox, Ellipse, RotatedBox]
        ]
    ] = Field(list())

    @validator("instances", pre=True, each_item=True)
    def check_instances(cls, instance):
        # todo add type checking
        annotation_type = instance.get("type")
        result = validate_model(ANNOTATION_TYPES[annotation_type], instance)
        if result[2]:
            raise ValidationError(
                result[2].raw_errors, model=ANNOTATION_TYPES[annotation_type]
            )
        return instance



from typing import Any
from typing import TypeVar

from pydantic import BaseModel
from pydantic import ValidationError
from pydantic.error_wrappers import ErrorWrapper


class CustomFieldMeta(type):
    def __getitem__(self, key_value_type: Any):
        return type('CustomFieldValue', (self,), {'key_value_type': key_value_type})


T = TypeVar("T")


#
# class Strict(Generic[T]):
#     __type_like__: T
#
#     @staticmethod
#     def _display_type(v: Any) -> str:
#         try:
#             return v.__name__
#         except AttributeError:
#             return str(v).replace("typing.", "")
#
#     @classmethod
#     def __class_getitem__(cls, type_like: T) -> T:
#         new_cls = new_class(
#             f"Strict[{cls._display_type(type_like)}]",
#             (cls,),
#             {},
#             lambda ns: ns.update({"__type_like__": type_like}),
#         )
#         return cast(T, new_cls)
#
#     @classmethod
#     def __get_validators__(cls) -> Generator[Callable[..., Any], None, None]:
#         yield cls.validate
#
#     @classmethod
#     def validate(cls, value: Any) -> T:
#         if not isinstancex(value, cls.__type_like__):
#             raise TypeError(f"{value!r} is not of valid type")
#         return value

class StrictType(BaseModel):
    class Config:
        key_field = "type"


class StrictDict(metaclass=CustomFieldMeta):
    key_value_type: Any = None

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):

        if not isinstance(v, dict):
            raise TypeError("type dict expected")

        key_type, value_type = cls.key_value_type
        errors = []

        for key in v:
            try:
                key_type.validate(key)
            except TypeError as e:
                errors.append(ErrorWrapper(exc=e, loc=""))

            try:
                value_type.validate((v[key]))
            except TypeError as e:
                errors.append(ErrorWrapper(exc=e, loc=""))

        if errors:
            raise ValidationError(errors=errors, model=cls)
        return v


class BaseDiscriminatedInstances(BaseModel):
    __root__: Any

    class Config:
        DISCRIMINATOR = "type"

    @classmethod
    def __get_validators__(cls):
        yield cls.return_instance

    @classmethod
    def return_instance(cls, values):
        try:
            discriminator = values[cls.Config.DISCRIMINATOR]
        except KeyError:
            raise TypeError(
                f"Missing required 'type' field for instance: {values}"
            )
        try:
            for instance in cls.__fields__["__root__"].type_.__dict__["__args__"]:
                if discriminator in instance.__fields__.get(cls.Config.DISCRIMINATOR).type_.__dict__["__args__"]:
                    return instance(**values)
            raise TypeError(f"Incorrect type: {discriminator}")
        except KeyError:
            raise TypeError(f"Incorrect type: {discriminator}")



class AppException(Exception):
    """
    Base exception for Licensee App. All exceptions thrown by inviter should
    extend this.
    """

    def __init__(self, message):
        super().__init__(message)

        self.message = message

    def __str__(self):
        return self.message


class InvalidInput(AppException):
    pass



DATE_REGEX = r"\d{4}-[01]\d-[0-3]\dT[0-2]\d:[0-5]\d:[0-5]\d(?:\.\d{3})Z"
DATE_TIME_FORMAT_ERROR_MESSAGE = "does not match expected format YYYY-MM-DDTHH:MM:SS.fffZ"
POINT_LABEL_VALUE_FORMAT_ERROR_MESSAGE = "str type expected"
POINT_LABEL_KEY_FORMAT_ERROR_MESSAGE = "does not match expected format ^[0-9]+$"
INVALID_DICT_MESSAGE = "value is not a valid dict"



from enum import Enum


class VectorAnnotationTypeEnum(str, Enum):
    BBOX = "bbox"
    ELLIPSE = "ellipse"
    TEMPLATE = "template"
    CUBOID = "cuboid"
    POLYLINE = "polyline"
    POLYGON = "polygon"
    POINT = "point"
    RBBOX = "rbbox"


class CreationTypeEnum(str, Enum):
    MANUAL = "Manual"
    PREDICTION = "Prediction"
    PRE_ANNOTATION = "Pre-annotation"


class AnnotationStatusEnum(str, Enum):
    NOT_STARTED = "NotStarted"
    IN_PROGRESS = "InProgress"
    QUALITY_CHECK = "QualityCheck"
    RETURNED = "Returned"
    COMPLETED = "Completed"
    SKIPPED = "Skipped"


class BaseRoleEnum(str, Enum):
    ADMIN = "Admin"
    ANNOTATOR = "Annotator"
    QA = "QA"


class ImageRoleEnum(str, Enum):
    CUSTOMER = "Customer"



from enum import Enum
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from superannotate_schemas.schemas.base import BaseAttribute
from superannotate_schemas.schemas.base import BaseInstance
from superannotate_schemas.schemas.base import BboxPoints
from superannotate_schemas.schemas.base import BaseMetadata
from superannotate_schemas.schemas.base import PointLabels
from superannotate_schemas.schemas.base import Tag
from superannotate_schemas.schemas.base import AnnotationStatusEnum

from pydantic import BaseModel
from pydantic import StrictFloat
from pydantic import constr
from pydantic import Field
from pydantic import StrictBool
from pydantic import StrictStr
from pydantic import StrictInt


class VideoType(str, Enum):
    EVENT = "event"
    BBOX = "bbox"


class Attribute(BaseAttribute):
    id: StrictInt
    group_id: StrictInt = Field(alias="groupId")


class MetaData(BaseMetadata):
    width: Optional[StrictInt]
    height: Optional[StrictInt]
    status: Optional[AnnotationStatusEnum]
    duration: Optional[StrictInt]
    error: Optional[StrictBool]


class BaseTimeStamp(BaseModel):
    active: Optional[bool]
    attributes: Optional[Dict[constr(regex=r"^[-|+]$"), List[Attribute]]]  # noqa: F722


class BboxTimeStamp(BaseTimeStamp):
    points: Optional[BboxPoints]


class BaseVideoInstance(BaseInstance):
    id: Optional[StrictStr]
    type: VideoType
    locked: Optional[StrictBool]
    timeline: Dict[StrictFloat, BaseTimeStamp]


class BboxInstance(BaseVideoInstance):
    point_labels: Optional[PointLabels] = Field(alias="pointLabels")
    timeline: Dict[StrictFloat, BboxTimeStamp]


class EventInstance(BaseVideoInstance):
    pass


class VideoAnnotation(BaseModel):
    metadata: MetaData
    instances: Optional[List[Union[EventInstance, BboxInstance]]] = Field(list())
    tags: Optional[List[Tag]] = Field(list())



from enum import Enum
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from superannotate_schemas.schemas.enums import AnnotationStatusEnum
from superannotate_schemas.schemas.base import BaseAttribute
from superannotate_schemas.schemas.base import BaseInstance
from superannotate_schemas.schemas.base import BboxPoints
from superannotate_schemas.schemas.base import BaseMetadata
from superannotate_schemas.schemas.base import NotEmptyStr
from superannotate_schemas.schemas.base import PointLabels
from superannotate_schemas.schemas.base import Tag

from pydantic import BaseModel
from pydantic import StrictBool
from pydantic import conlist
from pydantic import Field
from pydantic import StrictInt
from pydantic import StrictStr


class Attribute(BaseAttribute):
    name: NotEmptyStr
    group_name: NotEmptyStr = Field(alias="groupName")


class VideoType(str, Enum):
    EVENT = "event"
    BBOX = "bbox"


class MetaData(BaseMetadata):
    width: Optional[StrictInt]
    height: Optional[StrictInt]
    status: Optional[AnnotationStatusEnum]
    duration: Optional[StrictInt]
    error: Optional[StrictBool]


class BaseTimeStamp(BaseModel):
    timestamp: StrictInt
    attributes: Optional[List[Attribute]]


class BboxTimeStamp(BaseTimeStamp):
    points: BboxPoints


class EventTimeStamp(BaseTimeStamp):
    pass


class InstanceMetadata(BaseInstance):
    type: VideoType
    start: StrictInt
    end: StrictInt

    class Config:
        fields = {"creation_type": {"exclude": True}}


class BBoxInstanceMetadata(InstanceMetadata):
    type: StrictStr = Field(VideoType.BBOX, const=True)
    point_labels: Optional[PointLabels] = Field(alias="pointLabels")


class EventInstanceMetadata(InstanceMetadata):
    type: StrictStr = Field(VideoType.EVENT, const=True)


class BaseVideoInstance(BaseModel):
    metadata: InstanceMetadata
    id: Optional[str]
    type: VideoType
    locked: Optional[bool]
    timeline: Dict[float, BaseTimeStamp]


class BaseParameter(BaseModel):
    start: StrictInt
    end: StrictInt


class BboxParameter(BaseParameter):
    timestamps: conlist(BboxTimeStamp, min_items=2)


class EventParameter(BaseParameter):
    timestamps: conlist(EventTimeStamp, min_items=2)


class BboxInstance(BaseModel):
    meta: BBoxInstanceMetadata
    parameters: conlist(BboxParameter, min_items=1)


class EventInstance(BaseModel):
    meta: EventInstanceMetadata
    parameters: conlist(EventParameter, min_items=1)


ANNOTATION_TYPES = {
    VideoType.BBOX: BboxInstance,
    VideoType.EVENT: EventInstance
}


class VideoAnnotation(BaseModel):
    metadata: MetaData
    instances: Optional[List[Union[EventInstance, BboxInstance]]] = Field(list())
    tags: Optional[List[Tag]] = Field(list())



from typing import List
from typing import Optional

from superannotate_schemas.schemas.base import BaseImageAnnotationInstance
from superannotate_schemas.schemas.base import BaseAttribute
from superannotate_schemas.schemas.base import BaseImageMetadata
from superannotate_schemas.schemas.base import NotEmptyStr
from superannotate_schemas.schemas.base import StrictStr
from superannotate_schemas.schemas.base import PixelColor
from superannotate_schemas.schemas.base import Tag
from superannotate_schemas.schemas.base import Comment

from pydantic import BaseModel
from pydantic import Field


class Attribute(BaseAttribute):
    name: NotEmptyStr
    group_name: NotEmptyStr = Field(alias="groupName")


class MetaData(BaseImageMetadata):
    is_segmented: Optional[bool] = Field(alias="isSegmented")
    is_predicted: Optional[bool] = Field(alias="isPredicted")


class AnnotationPart(BaseModel):
    color: PixelColor


class AnnotationInstance(BaseImageAnnotationInstance):
    parts: List[AnnotationPart]
    class_name: StrictStr = Field(alias="className")
    attributes: Optional[List[Attribute]] = Field(list())


class PixelAnnotation(BaseModel):
    metadata: MetaData
    instances: List[AnnotationInstance]
    tags: Optional[List[Tag]] = Field(list())
    comments: Optional[List[Comment]] = Field(list())



from superannotate_schemas.schemas.external.video import VideoAnnotation
from superannotate_schemas.schemas.external.document import DocumentAnnotation
from superannotate_schemas.schemas.external.pixel import PixelAnnotation
from superannotate_schemas.schemas.external.vector import VectorAnnotation

__all__ = [
    "DocumentAnnotation",
    "VideoAnnotation",
    "PixelAnnotation",
    "VectorAnnotation"
]



import os
import sys

WORKING_DIR = os.path.split(os.path.realpath(__file__))[0]
sys.path.append(WORKING_DIR)

from superannotate_schemas.validators import AnnotationValidators


__version__ = "1.0.0"


__all__ = [
    "__version__",
    "AnnotationValidators"
]




from superannotate_schemas.schemas.internal.pixel import PixelAnnotation
from superannotate_schemas.schemas.internal.vector import VectorAnnotation
from superannotate_schemas.schemas.internal.video import VideoAnnotation
from superannotate_schemas.schemas.internal.document import DocumentAnnotation


__all__ = [
    "DocumentAnnotation",
    "VideoAnnotation",
    "PixelAnnotation",
    "VectorAnnotation"
]


import copy

from abc import ABCMeta
from abc import abstractmethod
from typing import Any
from typing import Type

from superannotate_schemas.utils import wrap_error
from superannotate_schemas.schemas.external import PixelAnnotation as ExternalPixelAnnotation
from superannotate_schemas.schemas.external import VectorAnnotation as ExternalVectorAnnotation
from superannotate_schemas.schemas.external import VideoAnnotation as ExternalVideoAnnotation
from superannotate_schemas.schemas.external import DocumentAnnotation as ExternalDocumentAnnotation

from superannotate_schemas.schemas.internal import PixelAnnotation as InternalPixelAnnotation
from superannotate_schemas.schemas.internal import VectorAnnotation as InternalVectorAnnotation
from superannotate_schemas.schemas.internal import VideoAnnotation as InternalVideoAnnotation
from superannotate_schemas.schemas.internal import DocumentAnnotation as InternalDocumentAnnotation


from pydantic import BaseModel
from pydantic import Extra
from pydantic import ValidationError


class BaseValidator(metaclass=ABCMeta):
    MODEL: Type[BaseModel]

    def __init__(self, data: Any, allow_extra: bool = True):
        self.data = data
        self._validation_output = None
        self._extra = Extra.allow if allow_extra else Extra.forbid

    @classmethod
    def validate(cls, data: Any, extra=True):
        return cls.MODEL(**data)

    def _validate(self):
        model = copy.deepcopy(self.MODEL)
        model.Config.extra = self._extra
        self.data = model(**self.data).dict(by_alias=True, exclude_none=True)

    @abstractmethod
    def is_valid(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def generate_report(self) -> str:
        raise NotImplementedError


class BaseSchemaValidator(BaseValidator):
    MODEL = BaseModel()

    def is_valid(self) -> bool:
        try:
            self._validate()
        except ValidationError as e:
            self._validation_output = e
        return not bool(self._validation_output)

    def generate_report(self) -> str:
        return wrap_error(self._validation_output)


class ValidatorFactory:
    @staticmethod
    def _get_default_schema():
        return type('CopyOfB', BaseSchemaValidator.__bases__, dict(BaseSchemaValidator.__dict__))

    @staticmethod
    def get_validator(model: BaseModel):
        schema = ValidatorFactory._get_default_schema()
        schema.MODEL = model
        return schema

    def __class_getitem__(cls, item):
        return cls.get_validator(item)


class AnnotationValidators:
    VALIDATORS = {
        "pixel": (
            ValidatorFactory[ExternalPixelAnnotation],
            ValidatorFactory[InternalPixelAnnotation]
        ),
        "vector": (
            ValidatorFactory[ExternalVectorAnnotation],
            ValidatorFactory[InternalVectorAnnotation]
        ),
        "video": (
            ValidatorFactory[ExternalVideoAnnotation],
            ValidatorFactory[InternalVideoAnnotation]
        ),
        "document": (
            ValidatorFactory[ExternalDocumentAnnotation],
            ValidatorFactory[InternalDocumentAnnotation]
        )
    }

    @classmethod
    def get_validator(cls, project_type: str, internal=False):
        idx = 1 if internal else 0
        return cls.VALIDATORS[project_type][idx]




from typing import List
from typing import Optional
from typing import Union

from superannotate_schemas.schemas.base import BaseVectorInstance
from superannotate_schemas.schemas.base import BboxPoints
from superannotate_schemas.schemas.base import BaseAttribute
from superannotate_schemas.schemas.base import Comment
from superannotate_schemas.schemas.base import BaseImageMetadata as Metadata
from superannotate_schemas.schemas.base import Tag
from superannotate_schemas.schemas.base import AxisPoint
from superannotate_schemas.schemas.base import VectorAnnotationTypeEnum
from superannotate_schemas.schemas.base import StrictNumber

from pydantic import BaseModel
from pydantic import StrictInt
from pydantic import StrictFloat
from pydantic import conlist
from pydantic import Field
from pydantic import validate_model
from pydantic import ValidationError
from pydantic import validator


class Attribute(BaseAttribute):
    id: StrictInt
    group_id: StrictInt = Field(alias="groupId")


class VectorInstance(BaseVectorInstance):
    attributes: Optional[List[Attribute]] = Field(list())


class Point(VectorInstance, AxisPoint):
    pass


class PolyLine(VectorInstance):
    points: List[Union[StrictFloat, StrictInt]]


class Polygon(VectorInstance):
    points: conlist(Union[StrictFloat, StrictInt], min_items=3)


class Bbox(VectorInstance):
    points: BboxPoints


class RotatedBoxPoints(VectorInstance):
    x1: StrictNumber
    y1: StrictNumber
    x2: StrictNumber
    y2: StrictNumber
    x3: StrictNumber
    y3: StrictNumber
    x4: StrictNumber
    y4: StrictNumber


class RotatedBox(VectorInstance):
    points: RotatedBoxPoints


class Ellipse(VectorInstance):
    cx: StrictNumber
    cy: StrictNumber
    rx: StrictNumber
    ry: StrictNumber
    angle: StrictNumber


class TemplatePoint(BaseModel):
    id: StrictInt
    x: StrictNumber
    y: StrictNumber


class TemplateConnection(BaseModel):
    id: StrictInt
    from_connection: StrictInt = Field(alias="from")
    to_connection: StrictInt = Field(alias="to")


class Template(VectorInstance):
    points: conlist(TemplatePoint, min_items=1)
    connections: List[TemplateConnection]
    template_id: StrictInt = Field(alias="templateId")


class CuboidPoint(BaseModel):
    f1: AxisPoint
    f2: AxisPoint
    r1: AxisPoint
    r2: AxisPoint


class Cuboid(VectorInstance):
    points: CuboidPoint


ANNOTATION_TYPES = {
    VectorAnnotationTypeEnum.BBOX: Bbox,
    VectorAnnotationTypeEnum.TEMPLATE: Template,
    VectorAnnotationTypeEnum.CUBOID: Cuboid,
    VectorAnnotationTypeEnum.POLYGON: Polygon,
    VectorAnnotationTypeEnum.POINT: Point,
    VectorAnnotationTypeEnum.POLYLINE: PolyLine,
    VectorAnnotationTypeEnum.ELLIPSE: Ellipse,
    VectorAnnotationTypeEnum.RBBOX: RotatedBox,
}


class VectorAnnotation(BaseModel):
    metadata: Metadata
    comments: Optional[List[Comment]] = Field(list())
    tags: Optional[List[Tag]] = Field(list())
    instances: Optional[
        List[
            Union[Template, Cuboid, Point, PolyLine, Polygon, Bbox, Ellipse, RotatedBox]
        ]
    ] = Field(list())

    @validator("instances", pre=True, each_item=True)
    def check_instances(cls, instance):
        # todo add type checking
        annotation_type = instance.get("type")
        result = validate_model(ANNOTATION_TYPES[annotation_type], instance)
        if result[2]:
            raise ValidationError(
                result[2].raw_errors, model=ANNOTATION_TYPES[annotation_type]
            )
        return instance



from typing import List
from typing import Optional

from superannotate_schemas.schemas.base import BaseImageMetadata as Metadata
from superannotate_schemas.schemas.base import PixelColor
from superannotate_schemas.schemas.base import BaseAttribute
from superannotate_schemas.schemas.base import BaseImageAnnotationInstance
from superannotate_schemas.schemas.base import Tag
from superannotate_schemas.schemas.base import Comment

from pydantic import BaseModel
from pydantic import StrictInt
from pydantic import Field


class Attribute(BaseAttribute):
    id: StrictInt
    group_id: StrictInt = Field(alias="groupId")


class AnnotationPart(BaseModel):
    color: PixelColor


class AnnotationInstance(BaseImageAnnotationInstance):
    parts: List[AnnotationPart]
    class_id: StrictInt = Field(alias="classId")
    attributes: Optional[List[Attribute]] = Field(list())


class PixelAnnotation(BaseModel):
    metadata: Metadata
    instances: List[AnnotationInstance]
    tags: Optional[List[Tag]] = Field(list())
    comments: Optional[List[Comment]] = Field(list())



from pathlib import Path
from setuptools import find_packages, setup

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


setup(
    name='superannotate_schemas',
    long_description=long_description,
    long_description_content_type='text/markdown',
    version='1.0.1',
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    description='SuperAnnotate JSON Schemas',
    author='Vaghinak Basentsyan',
    author_email='vaghinak@superannotate.con',
    url='https://www.superannotate.com/',
    license='MIT',
    description_file="README.md",
    entry_points={
            'console_scripts': ['superannotate_schemas = superannotate_schemas.bin.app:main']
        },
      )



from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from pydantic import BaseModel as PyDanticBaseModel
from pydantic import conlist
from pydantic import constr
from pydantic import EmailStr
from pydantic import Extra
from pydantic import StrictInt
from pydantic import StrictFloat
from pydantic import StrictStr
from pydantic import StrictBool
from pydantic import Field
from pydantic import StrRegexError
from pydantic import ValidationError
from pydantic.error_wrappers import ErrorWrapper
from pydantic import validator
from pydantic.validators import strict_str_validator
from pydantic.color import Color
from pydantic.color import ColorType

from superannotate_schemas.schemas.enums import CreationTypeEnum
from superannotate_schemas.schemas.enums import ImageRoleEnum
from superannotate_schemas.schemas.enums import VectorAnnotationTypeEnum
from superannotate_schemas.schemas.enums import AnnotationStatusEnum
from superannotate_schemas.schemas.constances import DATE_REGEX
from superannotate_schemas.schemas.constances import DATE_TIME_FORMAT_ERROR_MESSAGE
from superannotate_schemas.schemas.constances import POINT_LABEL_VALUE_FORMAT_ERROR_MESSAGE
from superannotate_schemas.schemas.constances import POINT_LABEL_KEY_FORMAT_ERROR_MESSAGE
from superannotate_schemas.schemas.constances import INVALID_DICT_MESSAGE


NotEmptyStr = constr(strict=True, min_length=1)

StrictNumber = Union[StrictInt, StrictFloat]


class BaseModel(PyDanticBaseModel):
    class Config:
        extra = Extra.allow
        use_enum_values = True
        error_msg_templates = {
            "type_error.integer": "integer type expected",
            "type_error.string": "str type expected",
            "value_error.missing": "field required",
        }


class AxisPoint(BaseModel):
    x: StrictNumber
    y: StrictNumber


class BaseAttribute(BaseModel):
    id: Optional[StrictInt]
    group_id: Optional[StrictInt] = Field(alias="groupId")
    name: Optional[NotEmptyStr]
    group_name: Optional[NotEmptyStr] = Field(alias="groupName")


class Tag(BaseModel):
    __root__: NotEmptyStr


class AttributeGroup(BaseModel):
    name: NotEmptyStr
    is_multiselect: Optional[bool] = Field(False)
    attributes: List[BaseAttribute]


class BboxPoints(BaseModel):
    x1: StrictNumber
    x2: StrictNumber
    y1: StrictNumber
    y2: StrictNumber


class TimedBaseModel(BaseModel):
    created_at: Optional[constr(regex=DATE_REGEX)] = Field(None, alias="createdAt")
    updated_at: Optional[constr(regex=DATE_REGEX)] = Field(None, alias="updatedAt")

    @validator("created_at", "updated_at", pre=True)
    def validate_created_at(cls, value):
        try:
            if value is not None:
                constr(regex=DATE_REGEX, strict=True).validate(value)
        except (TypeError, StrRegexError):
            raise TypeError(DATE_TIME_FORMAT_ERROR_MESSAGE)
        return value


class UserAction(BaseModel):
    email: EmailStr
    role: ImageRoleEnum


class TrackableModel(BaseModel):
    created_by: Optional[UserAction] = Field(None, alias="createdBy")
    updated_by: Optional[UserAction] = Field(None, alias="updatedBy")
    creation_type: Optional[CreationTypeEnum] = Field(
        CreationTypeEnum.PRE_ANNOTATION.value, alias="creationType"
    )

    @validator("creation_type", always=True)
    def clean_creation_type(cls, _):
        return CreationTypeEnum.PRE_ANNOTATION.value


class LastUserAction(BaseModel):
    email: EmailStr
    timestamp: StrictInt


class BaseInstance(TrackableModel, TimedBaseModel):
    class_id: Optional[StrictInt] = Field(None, alias="classId")
    class_name: Optional[NotEmptyStr] = Field(None, alias="className")


class BaseMetadata(BaseModel):
    name: NotEmptyStr
    url: Optional[StrictStr]
    status: Optional[AnnotationStatusEnum]
    annotator_email: Optional[EmailStr] = Field(None, alias="annotatorEmail")
    qa_email: Optional[EmailStr] = Field(None, alias="qaEmail")
    last_action: Optional[LastUserAction] = Field(None, alias="lastAction")
    project_id: Optional[StrictInt] = Field(None, alias="projectId")


class BaseImageMetadata(BaseMetadata):
    width: Optional[StrictInt]
    height: Optional[StrictInt]
    pinned: Optional[bool]


class Correspondence(BaseModel):
    text: NotEmptyStr
    email: EmailStr


class Comment(TimedBaseModel, TrackableModel):
    x: Union[StrictInt, StrictFloat]
    y: Union[StrictInt, StrictFloat]
    resolved: Optional[StrictBool] = Field(False)
    correspondence: conlist(Correspondence, min_items=1)


class BaseImageAnnotationInstance(BaseInstance):
    visible: Optional[StrictBool]
    locked: Optional[StrictBool]
    probability: Optional[StrictInt] = Field(100)
    attributes: Optional[List[BaseAttribute]] = Field(list())
    error: Optional[StrictBool]

    class Config:
        error_msg_templates = {
            "value_error.missing": "field required for annotation",
        }


class StringA(BaseModel):
    string: StrictStr


class PointLabels(BaseModel):
    __root__: Dict[constr(regex=r"^[0-9]+$", min_length=1,  strict=True), StrictStr]

    @classmethod
    def __get_validators__(cls):
        yield cls.validate_type
        yield cls.validate_value

    @validator("__root__", pre=True)
    def validate_value(cls, values):
        result = {}
        errors = []
        validate_key, validate_value = None, None
        for key, value in values.items():
            try:
                validate_key = constr(regex=r"^[0-9]+$", min_length=1,  strict=True).validate(key)
            except ValueError:
                errors.append(
                    ErrorWrapper(
                        ValueError(POINT_LABEL_KEY_FORMAT_ERROR_MESSAGE), str(key)
                    )
                )
            try:
                validate_value = strict_str_validator(value)
            except ValueError:
                errors.append(
                    ErrorWrapper(
                        ValueError(POINT_LABEL_VALUE_FORMAT_ERROR_MESSAGE), str(key)
                    )
                )
            if validate_key and validate_value:
                result.update({validate_key: validate_value})
        if errors:
            raise ValidationError(errors, cls)
        return result

    @classmethod
    def validate_type(cls, values):
        if not issubclass(type(values), dict):
            raise TypeError(INVALID_DICT_MESSAGE)
        return values


class BaseVectorInstance(BaseImageAnnotationInstance):
    type: VectorAnnotationTypeEnum
    point_labels: Optional[PointLabels] = Field(alias="pointLabels")
    tracking_id: Optional[str] = Field(alias="trackingId")
    group_id: Optional[int] = Field(alias="groupId")

#
# class Metadata(BaseMetadata):
#     name: NotEmptyStr
#     status: Optional[AnnotationStatusEnum]
#     pinned: Optional[StrictBool]
#     is_predicted: Optional[StrictBool] = Field(None, alias="isPredicted")


class PixelColor(BaseModel):
    __root__: ColorType

    @validator("__root__")
    def validate_color(cls, v):
        color = Color(v)
        return color.as_hex()



import os
import json
import errno
from pathlib import Path

from superannotate_schemas.schemas.external import PixelAnnotation as ExternalPixelAnnotation
from superannotate_schemas.schemas.external import VectorAnnotation as ExternalVectorAnnotation
from superannotate_schemas.schemas.external import VideoAnnotation as ExternalVideoAnnotation
from superannotate_schemas.schemas.external import DocumentAnnotation as ExternalDocumentAnnotation

from superannotate_schemas.schemas.internal import PixelAnnotation as InternalPixelAnnotation
from superannotate_schemas.schemas.internal import VectorAnnotation as InternalVectorAnnotation
from superannotate_schemas.schemas.internal import VideoAnnotation as InternalVideoAnnotation
from superannotate_schemas.schemas.internal import DocumentAnnotation as InternalDocumentAnnotation
from superannotate_schemas.exceptions import InvalidInput

from superannotate_schemas import __version__
from superannotate_schemas.utils import uniquify
from superannotate_schemas.validators import AnnotationValidators


class CLIInterface:
    """
    This is to validate Pixel, Vector, Image and Document annotations.
    """
    DEFAULT_PATH = "schemas/"
    EXTERNAL_SCHEMAS = (
        ExternalPixelAnnotation, ExternalVectorAnnotation, ExternalDocumentAnnotation, ExternalVideoAnnotation
    )
    INTERNAL_SCHEMAS = (
        InternalPixelAnnotation, InternalVectorAnnotation, InternalDocumentAnnotation, InternalVideoAnnotation
    )

    @staticmethod
    def _create_folder(path: str):
        if not os.path.exists(os.path.dirname(path)):
            try:
                os.makedirs(os.path.dirname(path))
            except OSError as exc:
                if exc.errno != errno.EEXIST:
                    raise

    @staticmethod
    def validate(*paths, project_type, internal=False, verbose=False, report_path=None):
        if not paths:
            raise InvalidInput("Please provide paths.")
        if project_type not in AnnotationValidators.VALIDATORS.keys():
            raise InvalidInput(
                f"Invalid project type, valid types are: {', '.join(AnnotationValidators.VALIDATORS.keys())}"
            )

        validator_class = AnnotationValidators.get_validator(project_type, internal)
        validation_result = []
        for path in paths:
            if Path(path).is_file():
                with open(path, "r") as file:
                    data = json.load(file)
                    validator = validator_class(data)
                    if not validator.is_valid():
                        report = validator.generate_report()
                        if verbose:
                            print(f"{'-'* 4}{path}\n{report}")
                        if report_path:
                            with open(f"{report_path}/{uniquify(Path(path).name)}") as validation_report:
                                validation_report.write(report)
                        else:
                            validation_result.append({path: False})
            else:
                print(f"Skip {path}")
        if not verbose:
            print(validation_result)

    @staticmethod
    def version():
        return f"Version : {__version__}"



import os
from collections import defaultdict

from pydantic import ValidationError


def get_tabulation_size() -> int:
    try:
        return int(os.get_terminal_size().columns / 2)
    except OSError:
        return 48


def wrap_error(e: ValidationError) -> str:
    tabulation = get_tabulation_size()
    error_messages = defaultdict(list)
    for error in e.errors():
        errors_list = list(error["loc"])
        if "__root__" in errors_list:
            errors_list.remove("__root__")
        errors_list[1::] = [
            f"[{i}]" if isinstance(i, int) else f".{i}" for i in errors_list[1::]
        ]
        error_messages["".join(errors_list)].append(error["msg"])
    texts = []
    for field, text in error_messages.items():
        texts.append(
            "{} {}{}".format(
                field,
                " " * (tabulation - len(field)),
                f"\n {' ' * tabulation}".join(text),
            )
        )
    return "\n".join(texts)


def uniquify(path):
    filename, extension = os.path.splitext(path)
    counter = 1

    while os.path.exists(path):
        path = filename + " (" + str(counter) + ")" + extension
        counter += 1

    return path


from typing import List
from typing import Optional

from superannotate_schemas.schemas.base import BaseAttribute
from superannotate_schemas.schemas.base import BaseInstance
from superannotate_schemas.schemas.base import BaseMetadata as Metadata
from superannotate_schemas.schemas.base import Tag
from superannotate_schemas.schemas.base import NotEmptyStr

from pydantic import BaseModel
from pydantic import Field
from pydantic import StrictStr


class Attribute(BaseAttribute):
    name: NotEmptyStr
    group_name: NotEmptyStr = Field(alias="groupName")


class DocumentInstance(BaseInstance):
    start: int
    end: int
    attributes: Optional[List[Attribute]] = Field(list())


class DocumentAnnotation(BaseModel):
    metadata: Metadata
    instances: Optional[List[DocumentInstance]] = Field(list())
    tags: Optional[List[Tag]] = Field(list())
    free_text: Optional[StrictStr] = Field(None, alias="freeText")
